/*
 * Copyright 2012-2017 Red Hat, Inc.
 *
 * This file is part of Thermostat.
 *
 * Thermostat is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2, or (at your
 * option) any later version.
 *
 * Thermostat is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Thermostat; see the file COPYING.  If not see
 * <http://www.gnu.org/licenses/>.
 *
 * Linking this code with other modules is making a combined work
 * based on this code.  Thus, the terms and conditions of the GNU
 * General Public License cover the whole combination.
 *
 * As a special exception, the copyright holders of this code give
 * you permission to link this code with independent modules to
 * produce an executable, regardless of the license terms of these
 * independent modules, and to copy and distribute the resulting
 * executable under terms of your choice, provided that you also
 * meet, for each linked independent module, the terms and conditions
 * of the license of that module.  An independent module is a module
 * which is not derived from or based on this code.  If you modify
 * this code, you may extend this exception to your version of the
 * library, but you are not obligated to do so.  If you do not wish
 * to do so, delete this exception statement from your version.
 */

package com.redhat.thermostat.common.yaml;

import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.google.gson.JsonPrimitive;

import java.io.IOException;
import java.io.StringWriter;
import java.io.Writer;
import java.util.Map;

/**
 * convert JSON object to YAML string
 */
public class JsonToYaml {

    private Writer out;
    private int indent = 2;

    /**
     * Convert JSON object to YAML string
     * @param root object to convert to YAML
     * @return the YAML string
     * @throws IOException in the unlikely event of a I/O error
     */
    String jsonToYamlString(JsonObject root) throws IOException {
        this.out = new StringWriter();
        jsonToYaml(root, out);
        return out.toString();
    }

    /**
     * Convert and write JSON object to YAML
     * @param root object to convert to YAML
     * @param out Writer to write YAML to
     * @throws IOException in the event of a I/O error
     */
    public void jsonToYaml(JsonObject root, Writer out) throws IOException {
        this.out = out;
        jsonToYaml(root, 0, false);
    }

    /**
     * set the indentation for YAML output
     * @param indent the number of spaces to indent nested elements by
     */
    public void setIndentation(int indent) {
        this.indent = indent;
    }

    private void dispatch(JsonElement el, int currentIndent) throws IOException {
        if (el.isJsonPrimitive()) {
            final JsonPrimitive pr = el.getAsJsonPrimitive();
            if (pr.isBoolean() || pr.isNumber()) {
                out.append(el.getAsString()).append('\n');
            } else {
                String escapedString = el.getAsString().replace("'", "''");
                out.append("'").append(escapedString).append("'\n");
            }
        } else if (el.isJsonArray()) {
            out.append('\n');
            jsonToYaml(el.getAsJsonArray(), currentIndent);
        } else if (el.isJsonObject()) {
            out.append('\n');
            jsonToYaml(el.getAsJsonObject(), currentIndent, false);
        } else {
            throw new UnsupportedOperationException("Unsupported element: " + el);
        }
    }

    private void jsonToYaml(JsonObject obj, int currentIndent, boolean skipFirstIndent) throws IOException {
        final int newindent = currentIndent + indent;
        boolean isFirst = true;
        for (final Map.Entry<String, JsonElement> entry : obj.entrySet()) {
            if (!(isFirst && skipFirstIndent)) {
                doIndent(currentIndent);
            }
            out.append(entry.getKey()).append(": ");
            dispatch(entry.getValue(), newindent);
            isFirst = false;
        }
    }

    private void jsonToYaml(JsonArray array, int currentIndent) throws IOException {
        final int newindent = currentIndent + indent;
        for (final JsonElement el : array) {
            doIndent(currentIndent);
            out.append("- ");
            if (el.isJsonArray()) {
                jsonToYaml(el.getAsJsonArray(), newindent);
            } else if (el.isJsonObject()) {
                jsonToYaml(el.getAsJsonObject(), newindent, true);
            } else {
                dispatch(el, newindent);
            }
        }
    }

    private void doIndent(int currentIndent) throws IOException {
        for (int i = 0; i < currentIndent; i++)
            out.append(' ');
    }
}

