/*
 * Copyright 2012-2017 Red Hat, Inc.
 *
 * This file is part of Thermostat.
 *
 * Thermostat is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2, or (at your
 * option) any later version.
 *
 * Thermostat is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Thermostat; see the file COPYING.  If not see
 * <http://www.gnu.org/licenses/>.
 *
 * Linking this code with other modules is making a combined work
 * based on this code.  Thus, the terms and conditions of the GNU
 * General Public License cover the whole combination.
 *
 * As a special exception, the copyright holders of this code give
 * you permission to link this code with independent modules to
 * produce an executable, regardless of the license terms of these
 * independent modules, and to copy and distribute the resulting
 * executable under terms of your choice, provided that you also
 * meet, for each linked independent module, the terms and conditions
 * of the license of that module.  An independent module is a module
 * which is not derived from or based on this code.  If you modify
 * this code, you may extend this exception to your version of the
 * library, but you are not obligated to do so.  If you do not wish
 * to do so, delete this exception statement from your version.
 */

package com.redhat.thermostat.common.yaml;

import com.google.gson.JsonArray;
import com.google.gson.JsonElement;
import com.google.gson.JsonObject;
import com.redhat.thermostat.common.json.JsonUtil;

import java.io.IOException;
import java.util.Map;

/**
 * Given a JSON object (created from a Swagger YAML or JSON file), resolve all $ref paths
 * TODO: currently doesn't test for circular references
 */
public class RefResolver {

    private static final String REF_MEMBER = "$ref";

    /**
     * resolve a path into an object
     * @param defsroot root of object that is referred to
     * @param current current object within defsroot (for relative paths)
     * @param path path to resolve (e.g. "#/definitions/return-code")
     * @return object pointed to by path
     * @throws IOException if path is unresolvable
     */
    public static JsonObject resolvePath(JsonObject defsroot, JsonObject current, String path) throws IOException {
        String[] pathElements = path.split("/");
        for (String pe : pathElements) {
            if ("#".equals(pe)) {
                current = defsroot;
            } else if (current.has(pe)) {
                current = current.getAsJsonObject(pe);
            } else {
                throw new IOException("unresolvable reference '" + pe + "' in path '" + path + "'");
            }
        }
        return current;
    }

    /**
     * replace all $ref paths by nested objects
     * @param defsRoot root of object that is referred to (usually the same as objeToResolve)
     * @param objToResolve root of object to resolve references of
     * @return objsToResolve
     * @throws IOException if any referencesa are unresolveable
     */
    public static JsonElement resolveRefs(JsonObject defsRoot, JsonElement objToResolve) throws IOException {

        if (objToResolve.isJsonObject()) {
            final JsonObject obj = objToResolve.getAsJsonObject();
            for (Map.Entry<String, JsonElement> entry : obj.entrySet()) {
                if (REF_MEMBER.equals(entry.getKey())) {
                    String ref = entry.getValue().getAsString();
                    obj.remove(REF_MEMBER);
                    JsonObject refObject = resolvePath(defsRoot, obj, ref);
                    resolveRefs(defsRoot, refObject);
                    // hoist all members of refObject to contents
                    for (Map.Entry<String, JsonElement> refentry : refObject.entrySet()) {
                        obj.add(refentry.getKey(), JsonUtil.deepCopy(refentry.getValue()));
                    }
                } else {
                    resolveRefs(defsRoot, entry.getValue());
                }
            }
        } else if (objToResolve.isJsonArray()) {
            final JsonArray contents = objToResolve.getAsJsonArray();
            for (int i = 0; i < contents.size(); i++) {
                JsonElement el = contents.get(i);
                if (el.isJsonObject()) {
                    contents.set(i, resolveRefs(defsRoot, el.getAsJsonObject()));
                }
            }
        }
        return objToResolve;
    }

}